import { isAsyncFunction, defineConfig } from '@agile-ts/utils';
import { extractRelevantObservers } from '../utils.js';
import { ComputedTracker } from './computed.tracker.js';
import { logCodeManager } from '../logCodeManager.js';
import { State } from '../state/state.js';

var __async = (__this, __arguments, generator) => {
  return new Promise((resolve, reject) => {
    var fulfilled = (value) => {
      try {
        step(generator.next(value));
      } catch (e) {
        reject(e);
      }
    };
    var rejected = (value) => {
      try {
        step(generator.throw(value));
      } catch (e) {
        reject(e);
      }
    };
    var step = (x) => x.done ? resolve(x.value) : Promise.resolve(x.value).then(fulfilled, rejected);
    step((generator = generator.apply(__this, __arguments)).next());
  });
};
class Computed extends State {
  constructor(agileInstance, computeFunction, config = {}) {
    super(agileInstance, Object.prototype.hasOwnProperty.call(config, "initialValue") ? config.initialValue : !isAsyncFunction(computeFunction) ? computeFunction() : null, {
      key: config.key,
      dependents: config.dependents
    });
    this.deps = new Set();
    this.hardCodedDeps = [];
    this.isComputed = true;
    this.computeFunction = computeFunction;
    config = defineConfig(config, {
      computedDeps: [],
      autodetect: !this.isComputeFunctionAsync
    });
    this.agileInstance = () => agileInstance;
    this.config = {
      autodetect: config.autodetect
    };
    this.hardCodedDeps = extractRelevantObservers(config.computedDeps).filter((dep) => dep !== void 0);
    this.deps = new Set(this.hardCodedDeps);
    this.deps.forEach((observer) => {
      observer.addDependent(this.observers["value"]);
    });
    this.recompute({ autodetect: config.autodetect, overwrite: true });
  }
  get computeFunction() {
    return this._computeFunction;
  }
  set computeFunction(value) {
    this._computeFunction = value;
    this.isComputeFunctionAsync = isAsyncFunction(value);
  }
  computeSync(config = {}) {
    config = defineConfig(config, {
      autodetect: this.config.autodetect
    });
    if (config.autodetect)
      ComputedTracker.track();
    const computeFunction = this.computeFunction;
    const computedValue = computeFunction();
    if (config.autodetect) {
      const foundDeps = ComputedTracker.getTrackedObservers();
      this.deps.forEach((observer) => {
        if (!foundDeps.includes(observer) && !this.hardCodedDeps.includes(observer)) {
          this.deps.delete(observer);
          observer.removeDependent(this.observers["value"]);
        }
      });
      foundDeps.forEach((observer) => {
        if (!this.deps.has(observer)) {
          this.deps.add(observer);
          observer.addDependent(this.observers["value"]);
        }
      });
    }
    return computedValue;
  }
  computeAsync() {
    return __async(this, null, function* () {
      return this.computeFunction();
    });
  }
  compute() {
    return __async(this, arguments, function* (config = {}) {
      if (config.autodetect && this.isComputeFunctionAsync) {
        logCodeManager.log("19:02:00");
      }
      return this.isComputeFunctionAsync ? this.computeAsync() : this.computeSync(config);
    });
  }
  computeAndIngest(config = {}) {
    if (this.isComputeFunctionAsync) {
      this.computeAsync().then((result) => {
        this.observers["value"].ingestValue(result, config);
      });
    } else {
      const result = this.computeSync(config);
      this.observers["value"].ingestValue(result, config);
    }
  }
  recompute(config = {}) {
    config = defineConfig(config, {
      autodetect: false
    });
    this.computeAndIngest(config);
    return this;
  }
  updateComputeFunction(computeFunction, deps = [], config = {}) {
    config = defineConfig(config, {
      autodetect: this.config.autodetect
    });
    this.deps.forEach((observer) => {
      observer.removeDependent(this.observers["value"]);
    });
    this.hardCodedDeps = extractRelevantObservers(deps).filter((dep) => dep !== void 0);
    this.deps = new Set(this.hardCodedDeps);
    this.deps.forEach((observer) => {
      observer.addDependent(this.observers["value"]);
    });
    this.computeFunction = computeFunction;
    this.recompute(config);
    return this;
  }
}

export { Computed };
